#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import uuidutils

from ironic.common import exception
from ironic.common import inspection_rules
from ironic.common.inspection_rules import base
from ironic.common.inspection_rules import engine
from ironic.common.inspection_rules import utils
from ironic.common.inspection_rules import validation
from ironic.conductor import task_manager
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


class TestInspectionRules(db_base.DbTestCase):
    def setUp(self):
        super(TestInspectionRules, self).setUp()
        self.node = obj_utils.create_test_node(self.context,
                                               driver='fake-hardware',
                                               driver_info={},
                                               extra={})

        self.sensitive_fields = ['password', 'auth_token', 'bmc_password']
        self.test_data = {
            'username': 'testuser',
            'password': 'secret123',
            'nested': {
                'normal': 'value',
                'password': 'nested_secret'
            },
            'list_data': [
                {'name': 'item1', 'password': 'item1_secret'},
                {'name': 'item2', 'normal': 'value2'}
            ],
            'auth_token': 'abc123token'
        }

        self.inventory = {
            'cpu': {'count': 4, 'architecture': 'x86_64'},
            'memory': {'total': 8192, 'physical_mb': 8192},
            'interfaces': [
                {'name': 'eth0', 'mac_address': '2a:03:9c:53:4e:46'},
                {'name': 'eth1', 'mac_address': 'a2:67:c1:b8:c1:bd'}
            ],
            'disks': [
                {'name': '/dev/sda', 'size': 1000000, 'model': 'test-disk-1'},
                {'name': '/dev/sdb', 'size': 2000000, 'model': 'test-disk-2'}
            ],
            'bmc_address': '192.168.1.100',
            'bmc_password': 'secret'
        }
        self.plugin_data = {"plugin": "data", "logs": "test logs",
                            "password": "plugin_secret"}

        self.rule1 = obj_utils.create_test_inspection_rule(self.context)
        self.rule2 = obj_utils.create_test_inspection_rule(self.context)
        self.sensitive_rule = obj_utils.create_test_inspection_rule(
            self.context, sensitive=True)


@mock.patch('ironic.objects.InspectionRule.list', autospec=True)
class TestApplyRules(TestInspectionRules):
    def setUp(self):
        super(TestApplyRules, self).setUp()

    @mock.patch.object(engine, 'get_built_in_rules', autospec=True)
    @mock.patch.object(engine, 'check_conditions', autospec=True)
    @mock.patch.object(engine, 'apply_actions', autospec=True)
    def test_apply_rules_no_rules(self, mock_apply_actions,
                                  mock_check_conditions, mock_get_built_in,
                                  mock_list):
        mock_list.return_value = []
        mock_get_built_in.return_value = []

        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = engine.apply_rules(task, self.inventory,
                                        self.plugin_data, 'main')

        mock_list.assert_called_once_with(
            context=self.context,
            filters={'phase': 'main'})
        mock_get_built_in.assert_called_once()

        mock_check_conditions.assert_not_called()
        mock_apply_actions.assert_not_called()
        self.assertIsNone(result)

    @mock.patch.object(engine, 'get_built_in_rules', autospec=True)
    @mock.patch.object(engine, 'check_conditions', autospec=True)
    @mock.patch.object(engine, 'apply_actions', autospec=True)
    def test_apply_rules_success(self, mock_apply_actions,
                                 mock_check_conditions, mock_get_built_in,
                                 mock_list):
        rule1 = {'uuid': 'rule-1', 'priority': 100, 'conditions': [],
                 'actions': [{'op': 'set-attribute',
                              'args': {'path': 'a', 'value': 'b'}}]}
        rule2 = {'uuid': 'rule-2', 'priority': 50, 'conditions': [],
                 'actions': [
                     {'op': 'set-capability',
                      'args': {'name': 'boot_mode', 'value': 'uefi'}}]}
        mock_list.return_value = [rule1]
        mock_get_built_in.return_value = [rule2]
        mock_check_conditions.return_value = True

        with task_manager.acquire(self.context, self.node.uuid) as task:
            engine.apply_rules(task, self.inventory, self.plugin_data, 'main')

        mock_list.assert_called_once_with(
            context=self.context,
            filters={'phase': 'main'})

        mock_get_built_in.assert_called_once()
        self.assertEqual(2, mock_check_conditions.call_count)
        self.assertEqual(2, mock_apply_actions.call_count)

    @mock.patch.object(engine, 'get_built_in_rules', autospec=True)
    @mock.patch.object(engine, 'check_conditions', autospec=True)
    @mock.patch.object(engine, 'apply_actions', autospec=True)
    def test_apply_rules_some_conditions_pass(self, mock_apply_actions,
                                              mock_check_conditions,
                                              mock_get_built_in,
                                              mock_list):
        """Test that rules are skipped when conditions don't match."""
        mock_list.return_value = [self.rule1]
        mock_get_built_in.return_value = [self.rule2]

        mock_check_conditions.side_effect = [False, True]

        with task_manager.acquire(self.context, self.node.uuid) as task:
            engine.apply_rules(task, self.inventory, self.plugin_data, 'main')

        self.assertEqual(2, mock_check_conditions.call_count)
        mock_apply_actions.assert_called_once()

    @mock.patch.object(engine, 'get_built_in_rules', autospec=True)
    @mock.patch.object(engine, 'check_conditions', autospec=True)
    @mock.patch.object(engine, 'apply_actions', autospec=True)
    def test_apply_rules_all_conditions_fail(self, mock_apply_actions,
                                             mock_check_conditions,
                                             mock_get_built_in,
                                             mock_list):
        """Test that rules are skipped when conditions don't match."""
        mock_list.return_value = [self.rule1]
        mock_get_built_in.return_value = [self.rule2]

        mock_check_conditions.side_effect = [False, False]

        with task_manager.acquire(self.context, self.node.uuid) as task:
            engine.apply_rules(task, self.inventory, self.plugin_data, 'main')

        self.assertEqual(2, mock_check_conditions.call_count)
        mock_apply_actions.assert_not_called()

    @mock.patch.object(engine, 'LOG', autospec=True)
    @mock.patch.object(engine, 'get_built_in_rules', autospec=True)
    @mock.patch.object(engine, 'check_conditions', autospec=True)
    @mock.patch.object(engine, 'apply_actions', autospec=True)
    def test_apply_rules_ironic_exception(self, mock_apply_actions,
                                          mock_check_conditions,
                                          mock_get_built_in,
                                          mock_log, mock_list):
        """Test that IronicException is re-raised."""
        mock_list.return_value = [self.rule1, self.rule2]
        mock_get_built_in.return_value = []
        mock_check_conditions.return_value = True

        mock_apply_actions.side_effect = [
            exception.IronicException("Expected error"),
            {'plugin_data': {'updated': 'data'}}
        ]

        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(exception.IronicException,
                              engine.apply_rules, task, self.inventory,
                              self.plugin_data, 'main')

        mock_log.error.assert_called_once()

        self.assertEqual(1, mock_apply_actions.call_count)

    @mock.patch.object(utils, 'ShallowMaskDict', autospec=True)
    @mock.patch.object(engine, 'get_built_in_rules', autospec=True)
    @mock.patch.object(engine, 'check_conditions', autospec=True)
    @mock.patch.object(engine, 'apply_actions', autospec=True)
    def test_apply_rules_with_always_mask(self, mock_apply_actions,
                                          mock_check_conditions,
                                          mock_get_built_in,
                                          mock_masked_dict, mock_list):
        """Test apply_rules with mask_secrets='always'."""
        self.config(mask_secrets='always', group='inspection_rules')

        mock_list.return_value = [self.rule1]
        mock_get_built_in.return_value = [self.rule2]
        mock_check_conditions.return_value = True

        masked_inventory = mock.MagicMock()
        masked_plugin_data = mock.MagicMock()
        mock_masked_dict.side_effect = [masked_inventory, masked_plugin_data,
                                        mock.MagicMock(), mock.MagicMock()]

        with task_manager.acquire(self.context, self.node.uuid) as task:
            engine.apply_rules(task, self.inventory, self.plugin_data, 'main')

        mock_masked_dict.assert_has_calls([
            mock.call(self.inventory,
                      sensitive_fields=engine.SENSITIVE_FIELDS,
                      mask_enabled=True),
            mock.call(self.plugin_data,
                      sensitive_fields=engine.SENSITIVE_FIELDS,
                      mask_enabled=True),
            mock.call(self.inventory,
                      sensitive_fields=engine.SENSITIVE_FIELDS,
                      mask_enabled=True),
            mock.call(self.plugin_data,
                      sensitive_fields=engine.SENSITIVE_FIELDS,
                      mask_enabled=True)
        ])

    @mock.patch.object(utils, 'ShallowMaskDict', autospec=True)
    @mock.patch.object(engine, 'get_built_in_rules', autospec=True)
    @mock.patch.object(engine, 'check_conditions', autospec=True)
    @mock.patch.object(engine, 'apply_actions', autospec=True)
    def test_apply_rules_with_never_mask(self, mock_apply_actions,
                                         mock_check_conditions,
                                         mock_get_built_in, mock_masked_dict,
                                         mock_list):
        """Test apply_rules with mask_secrets='never'."""
        self.config(mask_secrets='never', group='inspection_rules')

        mock_list.return_value = [self.rule1]
        mock_get_built_in.return_value = [self.rule2]
        mock_check_conditions.return_value = True

        masked_inventory = mock.MagicMock()
        masked_plugin_data = mock.MagicMock()
        mock_masked_dict.side_effect = [masked_inventory, masked_plugin_data,
                                        mock.MagicMock(), mock.MagicMock()]

        with task_manager.acquire(self.context, self.node.uuid) as task:
            engine.apply_rules(task, self.inventory, self.plugin_data, 'main')

        mock_masked_dict.assert_has_calls([
            mock.call(self.inventory,
                      sensitive_fields=engine.SENSITIVE_FIELDS,
                      mask_enabled=False),
            mock.call(self.plugin_data,
                      sensitive_fields=engine.SENSITIVE_FIELDS,
                      mask_enabled=False),
            mock.call(self.inventory,
                      sensitive_fields=engine.SENSITIVE_FIELDS,
                      mask_enabled=False),
            mock.call(self.plugin_data,
                      sensitive_fields=engine.SENSITIVE_FIELDS,
                      mask_enabled=False)
        ])

    @mock.patch.object(utils, 'ShallowMaskDict', autospec=True)
    @mock.patch.object(engine, 'get_built_in_rules', autospec=True)
    @mock.patch.object(engine, 'check_conditions', autospec=True)
    @mock.patch.object(engine, 'apply_actions', autospec=True)
    def test_apply_rules_with_sensitive_mask(self, mock_apply_actions,
                                             mock_check_conditions,
                                             mock_get_built_in,
                                             mock_masked_dict, mock_list):
        """Test apply_rules with mask_secrets='sensitive'."""
        self.config(mask_secrets='sensitive', group='inspection_rules')

        mock_list.return_value = [self.rule1, self.sensitive_rule]
        mock_get_built_in.return_value = []
        mock_check_conditions.return_value = True

        masked_inventory1 = mock.MagicMock()
        masked_plugin_data1 = mock.MagicMock()
        masked_inventory2 = mock.MagicMock()
        masked_plugin_data2 = mock.MagicMock()
        mock_masked_dict.side_effect = [
            masked_inventory1, masked_plugin_data1,
            masked_inventory2, masked_plugin_data2
        ]

        with task_manager.acquire(self.context, self.node.uuid) as task:
            engine.apply_rules(task, self.inventory, self.plugin_data, 'main')

        mock_masked_dict.assert_has_calls([
            mock.call(self.inventory,
                      sensitive_fields=engine.SENSITIVE_FIELDS,
                      mask_enabled=True),
            mock.call(self.plugin_data,
                      sensitive_fields=engine.SENSITIVE_FIELDS,
                      mask_enabled=True),
            mock.call(self.inventory,
                      sensitive_fields=engine.SENSITIVE_FIELDS,
                      mask_enabled=False),
            mock.call(self.plugin_data,
                      sensitive_fields=engine.SENSITIVE_FIELDS,
                      mask_enabled=False)
        ])


class TestOperators(TestInspectionRules):
    def setUp(self):
        super(TestOperators, self).setUp()

    def test_operator_exceptions(self):
        """Test that operators raise proper exceptions for invalid inputs."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            # NetOperator with invalid subnet
            net_op = inspection_rules.operators.NetOperator()
            self.assertRaises(
                exception.InspectionRuleExecutionFailure,
                net_op, task, address='192.168.1.1', subnet='invalid-subnet'
            )

            # MatchesOperator with invalid regex
            matches_op = inspection_rules.operators.MatchesOperator()
            self.assertRaises(
                exception.InspectionRuleExecutionFailure,
                matches_op, task, value='test', regex='[unclosed'
            )

            # ContainsOperator with invalid regex
            contains_op = inspection_rules.operators.ContainsOperator()
            self.assertRaises(
                exception.InspectionRuleExecutionFailure,
                contains_op, task, value='test', regex='[unclosed'
            )

            # SimpleOperator with non-list values
            eq_op = inspection_rules.operators.EqOperator()
            self.assertRaises(
                exception.RuleConditionCheckFailure,
                eq_op, task, values="not-a-list"
            )

    def test_oneofoperator_edge_cases(self):
        """Test OneOfOperator with edge cases."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            op = inspection_rules.operators.OneOfOperator()

            self.assertFalse(op(task, value='test', values=[]))
            self.assertFalse(op(task, value=None, values=['a', 'b']))
            self.assertTrue(op(task, value='a', values=['a', 'b']))

    def test_is_true_false_operators_edge_cases(self):
        """Test IsTrueOperator and IsFalseOperator."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            true_op = inspection_rules.operators.IsTrueOperator()
            false_op = inspection_rules.operators.IsFalseOperator()

            self.assertTrue(true_op(task, value='yes'))
            self.assertTrue(true_op(task, value='TRUE'))
            self.assertFalse(true_op(task, value='no'))

            self.assertTrue(false_op(task, value='no'))
            self.assertTrue(false_op(task, value='FALSE'))
            self.assertFalse(false_op(task, value='yes'))

            self.assertTrue(true_op(task, value=1))
            self.assertTrue(true_op(task, value=0.1))
            self.assertFalse(true_op(task, value=0))

            self.assertTrue(false_op(task, value=0))
            self.assertFalse(false_op(task, value=1))

            self.assertFalse(true_op(task, value=None))
            self.assertTrue(false_op(task, value=None))

            self.assertFalse(true_op(task, value={}))
            self.assertFalse(true_op(task, value=[]))

    def test_operator_with_loop(self):
        """Test operator check_with_loop method."""
        eq_condition = {
            'op': 'eq',
            'args': {'values': [1, '{item}']},
            'loop': [1, 2, 3, 4],
            'multiple': 'any'
        }

        contains_condition = {
            'op': 'contains',
            'args': {'value': '{item}', 'regex': '4'},
            'loop': ['test4', 'value5', 'string6'],
            'multiple': 'any'
        }

        oneof_condition = {
            'op': 'one-of',
            'args': {'value': '{inventory[cpu][architecture]}',
                     'values': ['{item}']},
            'loop': ['x86_64', 'aarch64', 'ppc64le'],
            'multiple': 'any'
        }

        with task_manager.acquire(self.context, self.node.uuid) as task:
            eq_op = inspection_rules.operators.EqOperator()
            contains_op = inspection_rules.operators.ContainsOperator()
            oneof_op = inspection_rules.operators.OneOfOperator()

            # 'any' multiple (should return True)
            self.assertTrue(eq_op.check_with_loop(
                task, eq_condition, self.inventory, self.plugin_data))
            self.assertTrue(contains_op.check_with_loop(
                task, contains_condition, self.inventory, self.plugin_data))
            self.assertTrue(oneof_op.check_with_loop(
                task, oneof_condition, self.inventory, self.plugin_data))

            # 'all' multiple (should return False)
            eq_condition['multiple'] = 'all'
            contains_condition['multiple'] = 'all'
            oneof_condition['multiple'] = 'all'

            self.assertFalse(eq_op.check_with_loop(
                task, eq_condition, self.inventory, self.plugin_data))
            self.assertFalse(contains_op.check_with_loop(
                task, contains_condition, self.inventory, self.plugin_data))
            self.assertFalse(oneof_op.check_with_loop(
                task, oneof_condition, self.inventory, self.plugin_data))

            # 'first' multiple (should return True)
            eq_condition['multiple'] = 'first'
            contains_condition['multiple'] = 'first'
            oneof_condition['multiple'] = 'first'

            self.assertTrue(eq_op.check_with_loop(
                task, eq_condition, self.inventory, self.plugin_data))
            self.assertTrue(contains_op.check_with_loop(
                task, contains_condition, self.inventory, self.plugin_data))
            self.assertTrue(oneof_op.check_with_loop(
                task, oneof_condition, self.inventory, self.plugin_data))

            # 'last' multiple (should return False for eq, True for others)
            eq_condition['multiple'] = 'last'
            contains_condition['multiple'] = 'last'
            oneof_condition['multiple'] = 'last'

            self.assertFalse(eq_op.check_with_loop(task, eq_condition,
                                                   self.inventory,
                                                   self.plugin_data))
            self.assertFalse(contains_op.check_with_loop(
                task, contains_condition, self.inventory, self.plugin_data))
            # This should be False since 'ppc64le' doesn't match 'x86_64'
            self.assertFalse(oneof_op.check_with_loop(
                task, oneof_condition, self.inventory, self.plugin_data))

    def test_rule_operators(self):
        """Test all inspection_rules.operators with True and False cases."""
        operator_tests = {
            inspection_rules.operators.EqOperator: [
                {'values': [5, 5]},
                {'values': [5, 10]}
            ],
            inspection_rules.operators.NeOperator: [
                {'values': [5, 10]},
                {'values': [5, 5]}
            ],
            inspection_rules.operators.LtOperator: [
                {'values': [5, 10]},
                {'values': [10, 5]}
            ],
            inspection_rules.operators.LeOperator: [
                {'values': [5, 10]},
                {'values': [10, 5]}
            ],
            inspection_rules.operators.GtOperator: [
                {'values': [10, 5]},
                {'values': [5, 10]}
            ],
            inspection_rules.operators.GeOperator: [
                {'values': [10, 5]},
                {'values': [5, 10]}
            ],
            inspection_rules.operators.EmptyOperator: [
                {'value': ''},
                {'value': 'not empty'}
            ],
            inspection_rules.operators.NetOperator: [
                {'address': '192.168.1.5', 'subnet': '192.168.1.0/24'},
                {'address': '10.0.0.1', 'subnet': '192.168.1.0/24'}
            ],
            inspection_rules.operators.MatchesOperator: [
                {'value': 'abc123', 'regex': r'abc\d+'},
                {'value': 'xyz123', 'regex': r'abc\d+'}
            ],
            inspection_rules.operators.ContainsOperator: [
                {'value': 'test-abc123-end', 'regex': r'abc\d+'},
                {'value': 'test-xyz-end', 'regex': r'abc\d+'}
            ],
            inspection_rules.operators.OneOfOperator: [
                {'value': 'b', 'values': ['a', 'b', 'c']},
                {'value': 'z', 'values': ['a', 'b', 'c']}
            ],
            inspection_rules.operators.IsNoneOperator: [
                {'value': None},
                {'value': 'something'}
            ],
            inspection_rules.operators.IsTrueOperator: [
                {'value': True},
                {'value': False}
            ],
            inspection_rules.operators.IsFalseOperator: [
                {'value': False},
                {'value': True}
            ]
        }

        with task_manager.acquire(self.context, self.node.uuid) as task:
            for op_class, test_cases in operator_tests.items():
                op = op_class()
                result = op(task, **test_cases[0])
                self.assertTrue(result)

                result = op(task, **test_cases[1])
                self.assertFalse(result)


class TestActions(TestInspectionRules):
    """Test inspection rule actions"""
    def setUp(self):
        super(TestActions, self).setUp()

    @mock.patch.object(inspection_rules.actions.LOG, 'info', autospec=True)
    def test_log_action(self, mock_log):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            action = inspection_rules.actions.LogAction()
            test_msg = "Test log message"
            action(task, msg=test_msg)
            mock_log.assert_called_once_with(test_msg)

    def test_fail_action(self):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            action = inspection_rules.actions.FailAction()
            error_msg = "Test failure"
            self.assertRaises(exception.HardwareInspectionFailure,
                              action, task, msg=error_msg)

    def test_action_path_dot_slash_notation(self):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            action = inspection_rules.actions.SetAttributeAction()

            # slash notation
            action(
                task, path='driver_info/new', value={'new_key': 'test_value'})

            # dot notation
            action(task, path='driver_info.next.nested.deeper',
                   value={'next_key': 'test_value'})

            self.assertEqual(
                {'new_key': 'test_value'}, task.node.driver_info['new'])
            self.assertEqual(
                {'nested': {'deeper': {'next_key': 'test_value'}}},
                task.node.driver_info['next'])

    def test_set_attribute_action(self):
        """Test SetAttributeAction sets node attribute."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            action = inspection_rules.actions.SetAttributeAction()

            action(
                task, path='driver_info/new', value={'new_key': 'test_value'})

            self.assertEqual(
                {'new_key': 'test_value'}, task.node.driver_info['new'])

    def test_extend_attribute_action(self):
        """Test ExtendAttributeAction extends a list attribute."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            action = inspection_rules.actions.ExtendAttributeAction()
            task.node.tags = ['existing']
            action(task, path='tags', value='new_tag')
            self.assertEqual(['existing', 'new_tag'], task.node.tags)

    def test_del_attribute_action(self):
        """Test DelAttributeAction deletes a node attribute."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            # Set up a value to delete
            task.node.extra = {'to_delete': 'value'}
            action = inspection_rules.actions.DelAttributeAction()
            action(task, path='extra/to_delete')
            self.assertEqual({}, task.node.extra)

    @mock.patch.object(inspection_rules.actions.objects.Trait, 'create',
                       autospec=True)
    def test_add_trait_action(self, mock_create):
        """Test AddTraitAction adds a node trait."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            action = inspection_rules.actions.AddTraitAction()
            trait_name = 'CUSTOM_AWESOME_TRAIT'
            action(task, name=trait_name)
            mock_create.assert_called_once()
            trait = mock_create.call_args[0][0]
            self.assertEqual(trait_name, trait.trait)

    @mock.patch.object(inspection_rules.actions.objects.Trait, 'destroy',
                       autospec=True)
    def test_remove_trait_action(self, mock_destroy):
        """Test RemoveTraitAction removes a node trait."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            action = inspection_rules.actions.RemoveTraitAction()
            trait_name = 'CUSTOM_AWESOME_TRAIT'
            action(task, name=trait_name)
            mock_destroy.assert_called_once_with(
                task.context, node_id=task.node.id, trait=trait_name)

    @mock.patch.object(inspection_rules.actions.driver_utils,
                       'add_node_capability', autospec=True)
    def test_set_capability_action(self, mock_add):
        """Test SetCapabilityAction sets a node capability."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            action = inspection_rules.actions.SetCapabilityAction()
            action(task, name='boot_mode', value='uefi')
            mock_add.assert_called_once_with(task, 'boot_mode', 'uefi')

    def test_unset_capability_action(self):
        """Test UnsetCapabilityAction removes a node capability."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.properties = {
                'capabilities': 'boot_mode:uefi,other:value'}
            action = inspection_rules.actions.UnsetCapabilityAction()
            action(task, name='boot_mode')
            self.assertEqual('other:value',
                             task.node.properties['capabilities'])

    def test_set_port_attribute_action(self):
        """Test SetPortAttributeAction sets a port attribute."""
        fake_port = mock.Mock()
        fake_port.uuid = uuidutils.generate_uuid()

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [fake_port]
            action = inspection_rules.actions.SetPortAttributeAction()
            action(task, port_id=fake_port.uuid, path='extra',
                   value='test_value')
            setattr(fake_port, 'extra', 'test_value')
            fake_port.save.assert_called_once()

    def test_extend_port_attribute_action(self):
        """Test ExtendPortAttributeAction extends a port attribute list."""
        fake_port = mock.Mock()
        fake_port.uuid = uuidutils.generate_uuid()
        fake_port.tags = ['existing']

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [fake_port]
            action = inspection_rules.actions.ExtendPortAttributeAction()
            action(task, port_id=fake_port.uuid, path='tags', value='new_tag')
            setattr(fake_port, 'tags', ['existing', 'new_tag'])
            fake_port.save.assert_called_once()

    def test_del_port_attribute_action(self):
        """Test DelPortAttributeAction deletes a port attribute."""
        fake_port = mock.Mock()
        fake_port.uuid = uuidutils.generate_uuid()
        fake_port.extra = {'to_delete': 'value'}

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.ports = [fake_port]
            action = inspection_rules.actions.DelPortAttributeAction()
            action(task, port_id=fake_port.uuid, path='extra/to_delete')
            fake_port.save.assert_called_once()

    def test_set_plugin_data_action(self):
        """Test SetPluginDataAction sets plugin data."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            plugin_data = {'existing': 'data'}
            action = inspection_rules.actions.SetPluginDataAction()
            action(task, path='test_key', value='test_value',
                   plugin_data=plugin_data)
            expected = {'existing': 'data', 'test_key': 'test_value'}
            self.assertEqual(expected, plugin_data)

    def test_extend_plugin_data_action(self):
        """Test ExtendPluginDataAction extends a plugin data list."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            plugin_data = {'test_list': ['item1']}
            action = inspection_rules.actions.ExtendPluginDataAction()
            action(task, path='test_list', value='item2',
                   plugin_data=plugin_data)
            expected = {'test_list': ['item1', 'item2']}
            self.assertEqual(expected, plugin_data)

    def test_unset_plugin_data_action(self):
        """Test UnsetPluginDataAction removes plugin data."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            plugin_data = {'to_remove': 'value', 'keep': 'value'}
            action = inspection_rules.actions.UnsetPluginDataAction()
            action(task, path='to_remove', plugin_data=plugin_data)
            self.assertEqual({'keep': 'value'}, plugin_data)

    def test_action_error_cases(self):
        """Test that actions properly handle error cases."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            # SetAttributeAction nested path on a non-dict
            set_attr = inspection_rules.actions.SetAttributeAction()
            task.node.driver = 'fake-hardware'
            self.assertRaises(
                exception.RuleActionExecutionFailure,
                set_attr, task, path='driver.some_key', value='test'
            )

            # ExtendAttributeAction non-list attribute
            task.node.driver = 'fake-hardware'
            extend_attr = inspection_rules.actions.ExtendAttributeAction()
            self.assertRaises(
                exception.RuleActionExecutionFailure,
                extend_attr, task, path='driver', value='new_item'
            )

            # DelAttributeAction nested path on a non-dict
            del_attr = inspection_rules.actions.DelAttributeAction()
            task.node.driver = 'fake-hardware'
            self.assertRaises(
                exception.RuleActionExecutionFailure,
                del_attr, task, path='driver.nonexistent_key'
            )

            # SetPortAttributeAction non-existent port
            set_port = inspection_rules.actions.SetPortAttributeAction()
            fake_port_id = uuidutils.generate_uuid()
            self.assertRaises(
                exception.PortNotFound,
                set_port, task, port_id=fake_port_id, path='extra',
                value='test'
            )

            # LogAction with invalid log level
            log_action = inspection_rules.actions.LogAction()
            self.assertRaises(
                exception.InspectionRuleExecutionFailure,
                log_action, task, msg='test message', level='invalid_level'
            )

    def test_action_with_list_loop(self):
        """Test action execute_with_loop method."""
        list_loop_data = {
            'op': 'set-attribute',
            'args': {'path': '{item[path]}', 'value': '{item[value]}'},
            'loop': [
                {'path': 'driver_info/ipmi_username', 'value': 'cidadmin'},
                {'path': 'driver_info/ipmi_password', 'value': 'cidpassword'},
                {
                    'path': 'driver_info/ipmi_address',
                    'value': '{inventory[bmc_address]}'
                }
            ]
        }

        with task_manager.acquire(self.context, self.node.uuid) as task:
            action = inspection_rules.actions.SetAttributeAction()
            action.execute_with_loop(task, list_loop_data, self.inventory,
                                     self.plugin_data)
            self.assertEqual('cidadmin',
                             task.node.driver_info['ipmi_username'])
            self.assertEqual('cidpassword',
                             task.node.driver_info['ipmi_password'])
            self.assertEqual('192.168.1.100',
                             task.node.driver_info['ipmi_address'])

    def test_action_with_dict_loop(self):
        """Test action execute_with_loop method."""
        dict_loop_data = {
            'op': 'set-attribute',
            'args': {'path': '{item[path]}', 'value': '{item[value]}'},
            'loop': {
                'path': 'driver_info/ipmi_address',
                'value': '{inventory[bmc_address]}'
            }
        }

        with task_manager.acquire(self.context, self.node.uuid) as task:
            action = inspection_rules.actions.SetAttributeAction()
            action.execute_with_loop(task, dict_loop_data, self.inventory,
                                     self.plugin_data)

            self.assertEqual('192.168.1.100',
                             task.node.driver_info['ipmi_address'])


class TestShallowMask(TestInspectionRules):
    def setUp(self):
        super(TestShallowMask, self).setUp()

    def test_set_mask_enabled(self):
        """Test that set_mask_enabled properly toggles masking."""
        masked_dict = utils.ShallowMaskDict(
            self.test_data, sensitive_fields=self.sensitive_fields,
            mask_enabled=True)

        self.assertEqual('***', masked_dict['password'])

        masked_dict.set_mask_enabled(False)
        self.assertEqual('secret123', masked_dict['password'])

        masked_dict.set_mask_enabled(True)
        self.assertEqual('***', masked_dict['password'])

    def test_getitem_masked(self):
        """Test that __getitem__ masks sensitive fields."""
        masked_dict = utils.ShallowMaskDict(
            self.test_data, sensitive_fields=self.sensitive_fields,
            mask_enabled=True)

        self.assertEqual('***', masked_dict['password'])
        self.assertEqual('***', masked_dict['auth_token'])

        self.assertEqual('testuser', masked_dict['username'])

    def test_getitem_not_masked(self):
        """Test that __getitem__ doesn't mask when disabled."""
        masked_dict = utils.ShallowMaskDict(
            self.test_data, sensitive_fields=self.sensitive_fields,
            mask_enabled=False)

        self.assertEqual('secret123', masked_dict['password'])
        self.assertEqual('abc123token', masked_dict['auth_token'])

    def test_nested_dict_masking(self):
        """Test that nested dictionaries are properly masked."""
        masked_dict = utils.ShallowMaskDict(
            self.test_data, sensitive_fields=self.sensitive_fields,
            mask_enabled=True)

        nested = masked_dict['nested']
        self.assertIsInstance(nested, utils.ShallowMaskDict)
        self.assertEqual('***', nested['password'])

        self.assertEqual('value', nested['normal'])

    def test_list_masking(self):
        """Test that lists containing dictionaries are properly masked."""
        masked_dict = utils.ShallowMaskDict(
            self.test_data, sensitive_fields=self.sensitive_fields,
            mask_enabled=True)

        list_data = masked_dict['list_data']
        self.assertEqual('***', list_data[0]['password'])

        self.assertEqual('item1', list_data[0]['name'])
        self.assertEqual('value2', list_data[1]['normal'])

    def test_items_masked(self):
        """Test that items() method returns masked values."""
        masked_dict = utils.ShallowMaskDict(
            self.test_data, sensitive_fields=self.sensitive_fields,
            mask_enabled=True)

        items = dict(masked_dict.items())

        self.assertEqual('***', items['password'])
        self.assertEqual('***', items['auth_token'])
        self.assertEqual('***', items['nested']['password'])

    def test_values_masked(self):
        """Test that values() method masks sensitive values."""
        test_data = {'username': 'user', 'password': 'secret'}
        masked_dict = utils.ShallowMaskDict(
            test_data, sensitive_fields=self.sensitive_fields,
            mask_enabled=True)

        values = list(masked_dict.values())

        self.assertIn('user', values)
        self.assertIn('***', values)
        self.assertNotIn('secret', values)

    def test_get_method_masked(self):
        """Test that the get() method properly masks sensitive fields."""
        masked_dict = utils.ShallowMaskDict(
            self.test_data, sensitive_fields=self.sensitive_fields,
            mask_enabled=True)

        self.assertEqual('***', masked_dict.get('password'))
        self.assertEqual('testuser', masked_dict.get('username'))

        # Non-existent field should return default
        self.assertIsNone(masked_dict.get('nonexistent'))
        self.assertEqual('default', masked_dict.get('nonexistent', 'default'))

    def test_modifying_dict(self):
        """Test that modifications affect the original data."""
        original_data = {'username': 'user', 'data': [1, 2, 3]}
        masked_dict = utils.ShallowMaskDict(
            original_data, sensitive_fields=self.sensitive_fields,
            mask_enabled=True)

        masked_dict['new_key'] = 'new_value'
        masked_dict['data'].append(4)

        self.assertEqual('new_value', original_data['new_key'])
        self.assertEqual([1, 2, 3, 4], original_data['data'])


class TestInterpolation(TestInspectionRules):
    def setUp(self):
        super(TestInterpolation, self).setUp()

    def test_variable_interpolation(self):
        """Test variable interpolation."""
        loop_context = {
            'item': {
                'key': 'value',
                'nested': {'deep': 'nested_value'}
            }
        }

        with task_manager.acquire(self.context, self.node.uuid) as task:
            value = "{inventory[cpu][architecture]}"
            result = base.Base.interpolate_variables(
                value, task.node, self.inventory, self.plugin_data)
            self.assertEqual("x86_64", result)

            value = "{inventory[interfaces][0][mac_address]}"
            result = base.Base.interpolate_variables(
                value, task.node, self.inventory, self.plugin_data)
            self.assertEqual("2a:03:9c:53:4e:46", result)

            value = "{plugin_data[plugin]}"
            result = base.Base.interpolate_variables(
                value, task.node, self.inventory, self.plugin_data)
            self.assertEqual("data", result)

            value = "{node.driver}"
            result = base.Base.interpolate_variables(
                value, task.node, self.inventory, self.plugin_data)
            self.assertEqual("fake-hardware", result)

            value = "{item}"
            result = base.Base.interpolate_variables(
                value, task.node, self.inventory,
                self.plugin_data, loop_context)

            self.assertEqual(
                "{'key': 'value', 'nested': {'deep': 'nested_value'}}",
                result)

            value = "{item[key]}"
            result = base.Base.interpolate_variables(
                value, task.node, self.inventory,
                self.plugin_data, loop_context)
            self.assertEqual("value", result)

            value = "{item[nested][deep]}"
            result = base.Base.interpolate_variables(
                value, task.node, self.inventory,
                self.plugin_data, loop_context)
            self.assertEqual("nested_value", result)

            value = "CPU: {inventory[cpu][count]}, Item: {item[key]}"
            result = base.Base.interpolate_variables(
                value, task.node, self.inventory,
                self.plugin_data, loop_context)
            self.assertEqual("CPU: 4, Item: value", result)

            dict_value = {
                "normal_key": "normal_value",
                "interpolated_key": "{inventory[cpu][architecture]}",
                "nested": {
                    "item_key": "{item[key]}",
                    "inventory_key": "{inventory[bmc_address]}"
                }
            }
            result = base.Base.interpolate_variables(
                dict_value, task.node, self.inventory,
                self.plugin_data, loop_context)
            self.assertEqual({
                "normal_key": "normal_value",
                "interpolated_key": "x86_64",
                "nested": {
                    "item_key": "value",
                    "inventory_key": "192.168.1.100"
                }
            }, result)

            list_value = [
                "normal_value",
                "{inventory[cpu][architecture]}",
                "{item[key]}",
                ["{inventory[bmc_address]}", "{item[nested][deep]}"]
            ]
            result = base.Base.interpolate_variables(
                list_value, task.node, self.inventory,
                self.plugin_data, loop_context)
            self.assertEqual([
                "normal_value",
                "x86_64",
                "value",
                ["192.168.1.100", "nested_value"]
            ], result)

            value = "{inventory[missing][key]}"
            result = base.Base.interpolate_variables(
                value, task.node, self.inventory, self.plugin_data)
            self.assertEqual(value, result)


class TestValidation(TestInspectionRules):
    def test_unsupported_operator_rejected(self):
        """Unsupported operator (even inverted) must raise Invalid."""
        rule = {
            'actions': [{'op': 'noop', 'args': {}}],
            'conditions': [{'op': '!unknown', 'args': {}}]
        }

        self.assertRaises(exception.Invalid, validation.validate_rule, rule)

    def test_conditions_not_list_raises_invalid(self):
        rule = {
            'actions': [{'op': 'noop', 'args': {}}],
            'conditions': {'op': 'eq', 'args': [1, 1]}  # not a list
        }

        self.assertRaises(exception.Invalid, validation.validate_rule, rule)

    def test_missing_actions_key_raises_invalid(self):
        rule = {
            'conditions': [{'op': 'eq', 'args': [1, 1]}]
            # 'actions' is missing
        }

        self.assertRaises(exception.Invalid, validation.validate_rule, rule)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
from unittest import mock

from oslo_config import cfg

from ironic.common import exception
from ironic.tests import base

CONF = cfg.CONF


class Unserializable(object):
    def __str__(self):
        raise NotImplementedError('nostr')


class TestException(exception.IronicException):
    _msg_fmt = 'Some exception: %(spam)s, %(ham)s'


class TestIronicException(base.TestCase):
    def test___init___json_serializable(self):
        exc = TestException(spam=[1, 2, 3], ham='eggs')
        self.assertIn('[1, 2, 3]', str(exc))
        self.assertEqual('[1, 2, 3]', exc.kwargs['spam'])

    def test___init___string_serializable(self):
        exc = TestException(
            spam=type('ni', (object,), dict(a=1, b=2))(), ham='eggs'
        )
        check_str = 'ni object at'
        self.assertIn(check_str, str(exc))
        self.assertIn(check_str, exc.kwargs['spam'])

    @mock.patch.object(exception.LOG, 'error', autospec=True)
    def test___init___invalid_kwarg(self, log_mock):
        CONF.set_override('fatal_exception_format_errors', False,
                          group='errors')
        e = TestException(spam=Unserializable(), ham='eggs')
        message = \
            log_mock.call_args_list[0][0][0] % log_mock.call_args_list[0][0][1]
        self.assertIsNotNone(
            re.search('spam: .*JSON.* NotImplementedError: nostr', message),
            message
        )
        self.assertEqual({'ham': '"eggs"', 'code': 500}, e.kwargs)

    @mock.patch.object(exception.LOG, 'error', autospec=True)
    def test___init___invalid_kwarg_reraise(self, log_mock):
        CONF.set_override('fatal_exception_format_errors', True,
                          group='errors')
        self.assertRaises(KeyError, TestException, spam=Unserializable(),
                          ham='eggs')
        message = \
            log_mock.call_args_list[0][0][0] % log_mock.call_args_list[0][0][1]
        self.assertIsNotNone(
            re.search('spam: .*JSON.* NotImplementedError: nostr', message),
            message
        )



# Copyright 2016 Intel Corporation

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from oslo_config import cfg

from ironic.common.i18n import _
from ironic.conf import auth


VALID_ADD_PORTS_VALUES = {
    'all': _('all MAC addresses'),
    'active': _('MAC addresses of NICs with IP addresses'),
    'pxe': _('only the MAC address of the PXE NIC'),
    'disabled': _('do not create any ports'),
}
VALID_KEEP_PORTS_VALUES = {
    'all': _('keep all ports, even ones with MAC addresses that are not '
             'present in the inventory'),
    'present': _('keep only ports with MAC addresses present in '
                 'the inventory'),
    'added': _('keep only ports determined by the add_ports option'),
}
DEFAULT_CPU_FLAGS_MAPPING = {
    'vmx': 'cpu_vt',
    'svm': 'cpu_vt',
    'aes': 'cpu_aes',
    'pse': 'cpu_hugepages',
    'pdpe1gb': 'cpu_hugepages_1g',
    'smx': 'cpu_txt',
}

opts = [
    cfg.IntOpt('status_check_period', default=60,
               help=_('period (in seconds) to check status of nodes '
                      'on inspection')),
    cfg.StrOpt('extra_kernel_params', default='',
               help=_('extra kernel parameters to pass to the inspection '
                      'ramdisk when boot is managed by ironic (not '
                      'ironic-inspector). Pairs key=value separated by '
                      'spaces.')),
    cfg.BoolOpt('power_off', default=True,
                help=_('whether to power off a node after inspection '
                       'finishes. Ignored for nodes that have fast '
                       'track mode enabled.')),
    cfg.StrOpt('callback_endpoint_override',
               help=_('endpoint to use as a callback for posting back '
                      'introspection data when boot is managed by ironic. '
                      'Standard keystoneauth options are used by default.')),
    # TODO(dtantsur): change the default to True when ironic-inspector is no
    # longer supported (and update the help string).
    cfg.BoolOpt('require_managed_boot', default=None,
                help=_('require that the in-band inspection boot is fully '
                       'managed by the node\'s boot interface. Set this to '
                       'False if your installation has a separate (i)PXE boot '
                       'environment for node discovery or unmanaged '
                       'inspection. You may need to set it to False to '
                       'inspect nodes that are not supported by boot '
                       'interfaces (e.g. because they don\'t have ports). '
                       'The default value depends on which inspect interface '
                       'is used: inspector uses False, agent - True.')),
    cfg.StrOpt('add_ports',
               default='pxe',
               help=_('Which MAC addresses to add as ports during '
                      'inspection.'),
               choices=list(VALID_ADD_PORTS_VALUES.items())),
    cfg.StrOpt('keep_ports',
               default='all',
               help=_('Which ports (already present on a node) to keep after '
                      'inspection.'),
               choices=list(VALID_KEEP_PORTS_VALUES.items())),
    cfg.BoolOpt('update_pxe_enabled',
                default=True,
                help=_('Whether to update the ports\' pxe_enabled field '
                       'according to the inspection data.')),
    cfg.StrOpt('default_hooks',
               default='ramdisk-error,validate-interfaces,ports,architecture',
               help=_('A comma-separated lists of inspection hooks that are '
                      'run by default for the "agent" inspection interface. '
                      'In most cases, the operators will not '
                      'modify this. The default (somewhat conservative) hooks '
                      'will raise an exception in case the ramdisk reports an '
                      'error, validate interfaces in the inventory, create '
                      'ports and set the node\'s cpu architecture property.')),
    cfg.StrOpt('hooks',
               default='$default_hooks',
               help=_('Comma-separated list of enabled hooks for processing '
                      'pipeline for the "agent" inspection interface. '
                      'The default for this is $default_hooks. '
                      'Hooks can be added before or after the defaults '
                      'like this: "prehook,$default_hooks,posthook".')),
    cfg.StrOpt('known_accelerators',
               default=os.path.join(
                   '$pybasedir',
                   'drivers/modules/inspector/hooks/known_accelerators.yaml'),
               help=_('Path to the file which contains the known accelerator '
                      'devices, to be used by the "accelerators" inspection '
                      'hook.')),
    cfg.DictOpt('cpu_capabilities',
                default=DEFAULT_CPU_FLAGS_MAPPING,
                help='Mapping between a CPU flag and a node capability to set '
                     'if this CPU flag is present. This configuration option '
                     'is used by the "cpu-capabilities" inspection hook.'),
    cfg.BoolOpt('extra_hardware_strict',
                default=False,
                help=_('If True, refuse to parse extra data (in plugin_data) '
                       'if at least one record is too short. Additionally, '
                       'remove the incoming "data" even if parsing failed. '
                       'This configuration option is used by the '
                       '"extra-hardware" inspection hook.')),
    cfg.MultiStrOpt('pci_device_alias',
                    default=[],
                    help=_('An alias for a PCI device identified by '
                           '\'vendor_id\' and \'product_id\' fields. Format: '
                           '{"vendor_id": "1234", "product_id": "5678", '
                           '"name": "pci_dev1"}. Use double quotes for the '
                           'keys and values.')),
    cfg.ListOpt('physical_network_cidr_map',
                default=[],
                sample_default=('10.10.10.0/24:physnet_a,'
                                '2001:db8::/64:physnet_b'),
                help=_('Mapping of IP subnet CIDR to physical network. When '
                       'the phyical-network inspection hook is enabled, the '
                       '"physical_network" property of corresponding '
                       'baremetal ports is populated based on this mapping.')),
    cfg.BoolOpt('disk_partitioning_spacing',
                default=True,
                help=_('Whether to leave 1 GiB of disk size untouched for '
                       'partitioning. Only has effect when used with the IPA '
                       'as a ramdisk, for older ramdisk local_gb is '
                       'calculated on the ramdisk side. This configuration '
                       'option is used by the "root-device" inspection hook.'))
]

discovery_opts = [
    cfg.BoolOpt('enabled',
                default=False, mutable=True,
                help=_("Setting this to True enables automatic enrollment "
                       "of inspected nodes that are not recognized. "
                       "When enabling this feature, keep in mind that any "
                       "machine hitting the inspection callback endpoint "
                       "will be automatically enrolled. The driver must be "
                       "set when setting this to True.")),
    cfg.StrOpt('driver',
               mutable=True,
               help=_("The default driver to use for newly enrolled nodes. "
                      "Must be set when enabling auto-discovery.")),
    cfg.StrOpt('inspection_scope',
               default=None,
               help=_("The default inspection scope for nodes enrolled via "
                      "auto-discovery.")),
]

pxe_filter_opts = [
    cfg.StrOpt('dhcp_hostsdir',
               help=_('The MAC address cache directory, exposed to dnsmasq.'
                      'This directory is expected to be in exclusive control '
                      'of the driver but must be purged by the operator. '
                      'Required.')),
    cfg.ListOpt('supported_inspect_interfaces',
                default=['agent'], mutable=True,
                help=_("List of inspect interfaces that will be considered "
                       "by the PXE filter. Only nodes with these interfaces "
                       "will be enabled.")),
    cfg.IntOpt('sync_period',
               default=45, mutable=True,
               help=_("Period (in seconds) between synchronizing the state "
                      "of dnsmasq with the database.")),
]

inspection_rule_opts = [
    cfg.StrOpt('built_in_rules',
               mutable=True,
               help=_("Path to YAML file of built-in inspection rules.")),
    cfg.StrOpt('mask_secrets',
               default='always',
               choices=['always', 'never', 'sensitive'],
               help=_("Whether to mask secrets in the node information "
                      "passed to the rules."))
]


def register_opts(conf):
    conf.register_opts(opts, group='inspector')
    conf.register_opts(discovery_opts, group='auto_discovery')
    conf.register_opts(pxe_filter_opts, group='pxe_filter')
    conf.register_opts(inspection_rule_opts, group='inspection_rules')
    auth.register_auth_opts(conf, 'inspector',
                            service_type='baremetal-introspection')


def list_opts():
    return auth.add_auth_opts(opts, service_type='baremetal-introspection')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from ironic.common import boot_devices
from ironic.common import exception
from ironic.drivers.modules import noop_mgmt
from ironic.tests import base


class TestNoopManagement(base.TestCase):
    iface = noop_mgmt.NoopManagement()

    def test_dummy_methods(self):
        self.assertEqual({}, self.iface.get_properties())
        self.assertIsNone(self.iface.validate("task"))
        self.assertEqual([boot_devices.PXE, boot_devices.DISK],
                         self.iface.get_supported_boot_devices("task"))
        self.assertEqual({'boot_device': boot_devices.PXE,
                          'persistent': True},
                         self.iface.get_boot_device("task"))

    def test_set_boot_device(self):
        self.iface.set_boot_device(mock.Mock(), boot_devices.DISK)
        self.assertRaises(exception.InvalidParameterValue,
                          self.iface.set_boot_device, mock.Mock(),
                          boot_devices.CDROM)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from ironic.common import driver_factory
from ironic.common import exception
from ironic.conductor import task_manager
from ironic.drivers import base as driver_base
from ironic.drivers.modules import agent
from ironic.drivers.modules import fake
from ironic.drivers.modules import inspector
from ironic.drivers.modules import noop
from ironic.drivers.modules import noop_mgmt
from ironic.drivers.modules import pxe
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


class ManualManagementHardwareTestCase(db_base.DbTestCase):

    def setUp(self):
        super(ManualManagementHardwareTestCase, self).setUp()
        self.config(enabled_hardware_types=['manual-management'],
                    enabled_power_interfaces=['fake'],
                    enabled_management_interfaces=['noop', 'fake'])

    def test_default_interfaces(self):
        node = obj_utils.create_test_node(self.context,
                                          driver='manual-management')
        with task_manager.acquire(self.context, node.id) as task:
            self.assertIsInstance(task.driver.management,
                                  noop_mgmt.NoopManagement)
            self.assertIsInstance(task.driver.power, fake.FakePower)
            self.assertIsInstance(task.driver.boot, pxe.PXEBoot)
            self.assertIsInstance(task.driver.deploy, agent.AgentDeploy)
            self.assertIsInstance(task.driver.inspect, noop.NoInspect)
            self.assertIsInstance(task.driver.raid, noop.NoRAID)

    def test_supported_interfaces(self):
        self.config(enabled_inspect_interfaces=['inspector', 'no-inspect'],
                    enabled_deploy_interfaces=['direct', 'custom-agent'],
                    enabled_raid_interfaces=['agent'])
        node = obj_utils.create_test_node(self.context,
                                          driver='manual-management',
                                          management_interface='fake',
                                          deploy_interface='custom-agent',
                                          raid_interface='agent')
        with task_manager.acquire(self.context, node.id) as task:
            self.assertIsInstance(task.driver.management, fake.FakeManagement)
            self.assertIsInstance(task.driver.power, fake.FakePower)
            self.assertIsInstance(task.driver.boot, pxe.PXEBoot)
            self.assertIsInstance(task.driver.deploy, agent.CustomAgentDeploy)
            self.assertIsInstance(task.driver.inspect, inspector.Inspector)
            self.assertIsInstance(task.driver.raid, agent.AgentRAID)

    def test_get_properties(self):
        # These properties are from vendor (agent) and boot (pxe) interfaces
        expected_prop_keys = [
            'agent_verify_ca', 'deploy_forces_oob_reboot',
            'deploy_kernel', 'deploy_ramdisk',
            'image_download_source', 'image_http_proxy',
            'image_https_proxy', 'image_no_proxy',
            'force_persistent_boot_device', 'kernel_append_params',
            'rescue_kernel', 'rescue_ramdisk']
        hardware_type = driver_factory.get_hardware_type("manual-management")
        properties = hardware_type.get_properties()
        self.assertCountEqual(expected_prop_keys, properties)

    @mock.patch.object(driver_factory, 'default_interface', autospec=True)
    def test_get_properties_none(self, mock_def_iface):
        hardware_type = driver_factory.get_hardware_type("manual-management")
        mock_def_iface.side_effect = exception.NoValidDefaultForInterface("no")
        properties = hardware_type.get_properties()
        self.assertEqual({}, properties)
        self.assertEqual(len(driver_base.ALL_INTERFACES),
                         mock_def_iface.call_count)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_context import context as oslo_context

from ironic.common import context
from ironic.tests import base as tests_base


class RequestContextTestCase(tests_base.TestCase):
    def setUp(self):
        super(RequestContextTestCase, self).setUp()
        self.context_dict = {
            'auth_token': 'auth_token1',
            "user_id": "user1",
            "project_id": "project1",
            "project_name": "somename",
            'read_only': True,
            'show_deleted': True,
            'request_id': 'id1',
            "is_public_api": True,
            "domain_id": "domain_id2",
            "user_domain_id": "domain_id3",
            "user_domain_name": "TreeDomain",
            "project_domain_id": "domain_id4",
            "roles": None,
            "overwrite": True
        }
        self.environ = {
            'keystone.token_info': {'key': 'value'},
        }

    @mock.patch.object(oslo_context.RequestContext, "__init__", autospec=True)
    def test_create_context(self, context_mock):
        test_context = context.RequestContext()
        context_mock.assert_called_once_with(mock.ANY)
        self.assertFalse(test_context.is_public_api)

    def test_to_policy_values(self):
        ctx = context.RequestContext(**self.context_dict)
        ctx_dict = ctx.to_policy_values()
        self.assertEqual('somename', ctx_dict['project_name'])
        self.assertTrue(ctx_dict['is_public_api'])

    @mock.patch.object(oslo_context, 'get_current', autospec=True)
    def test_thread_without_context(self, context_get_mock):
        self.context.update_store = mock.Mock()
        context_get_mock.return_value = None
        self.context.ensure_thread_contain_context()
        self.context.update_store.assert_called_once_with()

    @mock.patch.object(oslo_context, 'get_current', autospec=True)
    def test_thread_with_context(self, context_get_mock):
        self.context.update_store = mock.Mock()
        context_get_mock.return_value = self.context
        self.context.ensure_thread_contain_context()
        self.assertFalse(self.context.update_store.called)

    def test_create_context_with_environ(self):
        test_context = context.RequestContext().from_environ(self.environ)
        self.assertEqual({'key': 'value'}, test_context.auth_token_info)

    def test_to_dict_get_auth_token_info(self):
        test_context = context.RequestContext().from_environ(self.environ)
        self.assertEqual({'key': 'value'}, test_context.auth_token_info)
        result = test_context.to_dict()
        self.assertIn('auth_token_info', result)
        expected = {'key': 'value'}
        self.assertDictEqual(expected, result['auth_token_info'])

    def test_from_dict(self):
        cdict = {'auth_token_info': 'magicdict'}
        ctxt = context.RequestContext().from_dict(cdict)
        self.assertEqual('magicdict', ctxt.auth_token_info)

    def test_from_dict_older_api_server(self):
        # Validates auth_token_info is None if not included.
        cdict = {}
        ctxt = context.RequestContext().from_dict(cdict)
        self.assertIsNone(ctxt.auth_token_info)



#
# Copyright 2022 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from ironic.common.i18n import _

opts = [
    cfg.StrOpt('dhcp_optsdir',
               default='/etc/dnsmasq.d/optsdir.d',
               help=_('Directory where the "dnsmasq" provider will write '
                      'option configuration files for an external '
                      'Dnsmasq to read. Use the same path for the '
                      'dhcp-optsdir dnsmasq configuration directive.')),
    cfg.StrOpt('dhcp_hostsdir',
               default='/etc/dnsmasq.d/hostsdir.d',
               help=_('Directory where the "dnsmasq" provider will write '
                      'host configuration files for an external '
                      'Dnsmasq to read. Use the same path for the '
                      'dhcp-hostsdir dnsmasq configuration directive.')),
    cfg.StrOpt('dhcp_leasefile',
               default='/var/lib/dnsmasq/dnsmasq.leases',
               help=_('Dnsmasq leases file for the "dnsmasq" driver to '
                      'discover IP addresses of managed nodes. Use the'
                      'same path for the dhcp-leasefile dnsmasq '
                      'configuration directive.')),
]


def register_opts(conf):
    conf.register_opts(opts, group='dnsmasq')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client

from oslo_log import log
import pecan

from ironic.api.controllers.v1 import utils as api_utils
from ironic.api import method
from ironic.common import args
from ironic.common import exception
from ironic.common import metrics_utils

METRICS = metrics_utils.get_metrics_logger(__name__)

LOG = log.getLogger(__name__)


NETWORK_EVENT_VALIDATOR = args.and_valid(
    args.schema({
        'type': 'object',
        'properties': {
            'event': {'type': 'string'},
            'port_id': {'type': 'string'},
            'mac_address': {'type': 'string'},
            'status': {'type': 'string'},
            'device_id': {'type': ['string', 'null']},
            'binding:host_id': {'type': ['string', 'null']},
            'binding:vnic_type': {'type': ['string', 'null']},
        },
        'required': ['event', 'port_id', 'mac_address', 'status'],
        'additionalProperties': False,
    }),
    args.dict_valid(**{
        'port_id': args.uuid,
        'mac_address': args.mac_address,
        'device_id': args.uuid,
        'binding:host_id': args.uuid
    })
)

EVENT_VALIDATORS = {
    'network.bind_port': NETWORK_EVENT_VALIDATOR,
    'network.unbind_port': NETWORK_EVENT_VALIDATOR,
    'network.delete_port': NETWORK_EVENT_VALIDATOR,
}

EVENTS_SCHEMA = {
    'type': 'object',
    'properties': {
        'events': {
            'type': 'array',
            'minItems': 1,
            'items': {
                'type': 'object',
                'properties': {
                    'event': {'type': 'string',
                              'enum': list(EVENT_VALIDATORS)},
                },
                'required': ['event'],
                'additionalProperties': True,
            },
        },
    },
    'required': ['events'],
    'additionalProperties': False,
}


def events_valid(name, value):
    """Validator for events"""

    for event in value['events']:
        validator = EVENT_VALIDATORS[event['event']]
        validator(name, event)
    return value


class EventsController(pecan.rest.RestController):
    """REST controller for Events."""

    @pecan.expose()
    def _lookup(self):
        if not api_utils.allow_expose_events():
            pecan.abort(http_client.NOT_FOUND)

    @METRICS.timer('EventsController.post')
    @method.expose(status_code=http_client.NO_CONTENT)
    @method.body('evts')
    @args.validate(evts=args.and_valid(args.schema(EVENTS_SCHEMA),
                                       events_valid))
    def post(self, evts):
        if not api_utils.allow_expose_events():
            raise exception.NotFound()
        api_utils.check_policy('baremetal:events:post')
        for e in evts['events']:
            LOG.debug("Received external event: %s", e)



# coding=utf-8
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import versionutils
from oslo_versionedobjects import base as object_base

from ironic.db import api as dbapi
from ironic.objects import base
from ironic.objects import fields as object_fields


@base.IronicObjectRegistry.register
class BIOSSetting(base.IronicObject):
    # Version 1.0: Initial version
    # Version 1.1: Added registry
    VERSION = '1.1'

    dbapi = dbapi.get_instance()

    registry_fields = ('attribute_type', 'allowable_values', 'lower_bound',
                       'max_length', 'min_length', 'read_only',
                       'reset_required', 'unique', 'upper_bound')

    fields = {
        'node_id': object_fields.StringField(nullable=False),
        'name': object_fields.StringField(nullable=False),
        'value': object_fields.StringField(nullable=True),
        'attribute_type': object_fields.StringField(nullable=True),
        'allowable_values': object_fields.ListOfStringsField(
            nullable=True),
        'lower_bound': object_fields.IntegerField(nullable=True),
        'max_length': object_fields.IntegerField(nullable=True),
        'min_length': object_fields.IntegerField(nullable=True),
        'read_only': object_fields.BooleanField(nullable=True),
        'reset_required': object_fields.BooleanField(nullable=True),
        'unique': object_fields.BooleanField(nullable=True),
        'upper_bound': object_fields.IntegerField(nullable=True)
    }

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable
    def create(self, context=None):
        """Create a BIOS Setting record in DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: BIOSSetting(context)
        :raises: NodeNotFound if the node id is not found.
        :raises: BIOSSettingAlreadyExists if the setting record already exists.
        """
        values = self.do_version_changes_for_db()
        settings = {'name': values['name'], 'value': values['value']}
        for r in self.registry_fields:
            settings[r] = values.get(r)

        db_bios_setting = self.dbapi.create_bios_setting_list(
            values['node_id'], [settings], values['version'])
        self._from_db_object(self._context, self, db_bios_setting[0])

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable
    def save(self, context=None):
        """Save BIOS Setting update in DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: BIOSSetting(context)
        :raises: NodeNotFound if the node id is not found.
        :raises: BIOSSettingNotFound if the bios setting name is not found.
        """
        values = self.do_version_changes_for_db()

        settings = {'name': values['name'], 'value': values['value']}
        for r in self.registry_fields:
            settings[r] = values.get(r)

        updated_bios_setting = self.dbapi.update_bios_setting_list(
            values['node_id'], [settings], values['version'])
        self._from_db_object(self._context, self, updated_bios_setting[0])

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def get(cls, context, node_id, name):
        """Get a BIOS Setting based on its node_id and name.

        :param context: Security context.
        :param node_id: The node id.
        :param name: BIOS setting name to be retrieved.
        :raises: NodeNotFound if the node id is not found.
        :raises: BIOSSettingNotFound if the bios setting name is not found.
        :returns: A :class:'BIOSSetting' object.
        """
        db_bios_setting = cls.dbapi.get_bios_setting(node_id, name)
        return cls._from_db_object(context, cls(), db_bios_setting)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def delete(cls, context, node_id, name):
        """Delete a BIOS Setting based on its node_id and name.

        :param context: Security context.
        :param node_id: The node id.
        :param name: BIOS setting name to be deleted.
        :raises: NodeNotFound if the node id is not found.
        :raises: BIOSSettingNotFound if the bios setting name is not found.
        """
        cls.dbapi.delete_bios_setting_list(node_id, [name])

    def _convert_to_version(self, target_version,
                            remove_unavailable_fields=True):
        """Convert to the target version.

        Convert the object to the target version. The target version may be
        the same, older, or newer than the version of the object. This is
        used for DB interactions as well as for serialization/deserialization.

        Version 1.1: remove registry field for unsupported versions if
            remove_unavailable_fields is True.

        :param target_version: the desired version of the object
        :param remove_unavailable_fields: True to remove fields that are
            unavailable in the target version; set this to True when
            (de)serializing. False to set the unavailable fields to appropriate
            values; set this to False for DB interactions.
        """
        target_version = versionutils.convert_version_to_tuple(target_version)

        for field in self.get_registry_fields():
            field_is_set = self.obj_attr_is_set(field)
            if target_version >= (1, 1):
                # target version supports the major/minor specified
                if not field_is_set:
                    # set it to its default value if it is not set
                    setattr(self, field, None)
            elif field_is_set:
                # target version does not support the field, and it is set
                if remove_unavailable_fields:
                    # (De)serialising: remove unavailable fields
                    delattr(self, field)
                elif self.registry:
                    setattr(self, field, None)


@base.IronicObjectRegistry.register
class BIOSSettingList(base.IronicObjectListBase, base.IronicObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    dbapi = dbapi.get_instance()

    fields = {
        'objects': object_fields.ListOfObjectsField('BIOSSetting'),
    }

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def create(cls, context, node_id, settings):
        """Create a list of BIOS Setting records in DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: BIOSSetting(context)
        :param node_id: The node id.
        :param settings: A list of bios settings.
        :raises: NodeNotFound if the node id is not found.
        :raises: BIOSSettingAlreadyExists if any of the setting records
            already exists.
        :return: A list of BIOSSetting objects.
        """
        version = BIOSSetting.get_target_version()
        db_setting_list = cls.dbapi.create_bios_setting_list(
            node_id, settings, version)
        return object_base.obj_make_list(
            context, cls(), BIOSSetting, db_setting_list)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def save(cls, context, node_id, settings):
        """Save a list of BIOS Setting updates in DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: BIOSSetting(context)
        :param node_id: The node id.
        :param settings: A list of bios settings.
        :raises: NodeNotFound if the node id is not found.
        :raises: BIOSSettingNotFound if any of the bios setting names
            is not found.
        :return: A list of BIOSSetting objects.
        """
        version = BIOSSetting.get_target_version()
        updated_setting_list = cls.dbapi.update_bios_setting_list(
            node_id, settings, version)
        return object_base.obj_make_list(
            context, cls(), BIOSSetting, updated_setting_list)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def delete(cls, context, node_id, names):
        """Delete BIOS Settings based on node_id and names.

        :param context: Security context.
        :param node_id: The node id.
        :param names: List of BIOS setting names to be deleted.
        :raises: NodeNotFound if the node id is not found.
        :raises: BIOSSettingNotFound if any of BIOS setting fails to delete.
        """
        cls.dbapi.delete_bios_setting_list(node_id, names)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def get_by_node_id(cls, context, node_id):
        """Get BIOS Setting based on node_id.

        :param context: Security context.
        :param node_id: The node id.
        :raises: NodeNotFound if the node id is not found.
        :return: A list of BIOSSetting objects.
        """
        node_bios_setting = cls.dbapi.get_bios_setting_list(node_id)
        return object_base.obj_make_list(
            context, cls(), BIOSSetting, node_bios_setting)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def sync_node_setting(cls, context, node_id, settings):
        """Returns lists of create/update/delete/unchanged settings.

        This method sync with 'bios_settings' database table and sorts
        out four lists of create/update/delete/unchanged settings.

        :param context: Security context.
        :param node_id: The node id.
        :param settings: BIOS settings to be synced.
        :returns: A 4-tuple of lists of BIOS settings to be created,
            updated, deleted and unchanged.
        """
        create_list = []
        update_list = []
        delete_list = []
        nochange_list = []
        current_settings_dict = {}

        given_setting_names = [setting['name'] for setting in settings]
        current_settings = cls.get_by_node_id(context, node_id)

        for setting in current_settings:
            current_settings_dict[setting.name] = setting.value

        for setting in settings:
            if setting['name'] in current_settings_dict:
                if setting['value'] != current_settings_dict[setting['name']]:
                    update_list.append(setting)
                else:
                    nochange_list.append(setting)
            else:
                create_list.append(setting)

        for setting in current_settings:
            if setting.name not in given_setting_names:
                delete_list.append({'name': setting.name,
                                    'value': setting.value})

        return (create_list, update_list, delete_list, nochange_list)



# Copyright 2014 Rackspace Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import uuidutils

from ironic.common import exception
from ironic.common import network
from ironic.common import neutron as neutron_common
from ironic.common import states
from ironic.conductor import task_manager
from ironic.drivers.modules.network import common as driver_common
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.objects import utils as object_utils


class TestNetwork(db_base.DbTestCase):

    def setUp(self):
        super(TestNetwork, self).setUp()
        self.node = object_utils.create_test_node(self.context)

    def test_get_node_vif_ids_no_ports_no_portgroups(self):
        expected = {'portgroups': {},
                    'ports': {}}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = network.get_node_vif_ids(task)
        self.assertEqual(expected, result)

    def test_get_node_vif_ids_one_port_int_info(self):
        kwargs1 = {'internal_info': {'tenant_vif_port_id': 'test-vif-A'}}
        port1 = db_utils.create_test_port(node_id=self.node.id,
                                          address='aa:bb:cc:dd:ee:ff',
                                          uuid=uuidutils.generate_uuid(),
                                          **kwargs1)
        expected = {'portgroups': {},
                    'ports': {port1.uuid: 'test-vif-A'}}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = network.get_node_vif_ids(task)
        self.assertEqual(expected, result)

    def test_get_node_vif_ids_one_portgroup_int_info(self):
        kwargs1 = {'internal_info': {'tenant_vif_port_id': 'test-vif-A'}}
        pg1 = db_utils.create_test_portgroup(
            node_id=self.node.id, **kwargs1)

        expected = {'portgroups': {pg1.uuid: 'test-vif-A'},
                    'ports': {}}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = network.get_node_vif_ids(task)
        self.assertEqual(expected, result)

    def test_get_node_vif_ids_two_ports_int_info(self):
        kwargs1 = {'internal_info': {'tenant_vif_port_id': 'test-vif-A'}}
        kwargs2 = {'internal_info': {'tenant_vif_port_id': 'test-vif-B'}}
        port1 = db_utils.create_test_port(node_id=self.node.id,
                                          address='aa:bb:cc:dd:ee:ff',
                                          uuid=uuidutils.generate_uuid(),
                                          **kwargs1)
        port2 = db_utils.create_test_port(node_id=self.node.id,
                                          address='dd:ee:ff:aa:bb:cc',
                                          uuid=uuidutils.generate_uuid(),
                                          **kwargs2)
        expected = {'portgroups': {},
                    'ports': {port1.uuid: 'test-vif-A',
                              port2.uuid: 'test-vif-B'}}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = network.get_node_vif_ids(task)
        self.assertEqual(expected, result)

    def test_get_node_vif_ids_two_portgroups_int_info(self):
        kwargs1 = {'internal_info': {'tenant_vif_port_id': 'test-vif-A'}}
        kwargs2 = {'internal_info': {'tenant_vif_port_id': 'test-vif-B'}}
        pg1 = db_utils.create_test_portgroup(
            node_id=self.node.id, **kwargs1)
        pg2 = db_utils.create_test_portgroup(
            uuid=uuidutils.generate_uuid(),
            address='dd:ee:ff:aa:bb:cc',
            node_id=self.node.id,
            name='barname', **kwargs2)
        expected = {'portgroups': {pg1.uuid: 'test-vif-A',
                                   pg2.uuid: 'test-vif-B'},
                    'ports': {}}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = network.get_node_vif_ids(task)
        self.assertEqual(expected, result)

    def _test_get_node_vif_ids_multitenancy(self, int_info_key):
        port = db_utils.create_test_port(
            node_id=self.node.id, address='aa:bb:cc:dd:ee:ff',
            internal_info={int_info_key: 'test-vif-A'})
        portgroup = db_utils.create_test_portgroup(
            node_id=self.node.id, address='dd:ee:ff:aa:bb:cc',
            internal_info={int_info_key: 'test-vif-B'})
        expected = {'ports': {port.uuid: 'test-vif-A'},
                    'portgroups': {portgroup.uuid: 'test-vif-B'}}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = network.get_node_vif_ids(task)
        self.assertEqual(expected, result)

    def test_get_node_vif_ids_during_cleaning(self):
        self._test_get_node_vif_ids_multitenancy('cleaning_vif_port_id')

    def test_get_node_vif_ids_during_provisioning(self):
        self._test_get_node_vif_ids_multitenancy('provisioning_vif_port_id')

    def test_get_node_vif_ids_during_rescuing(self):
        self._test_get_node_vif_ids_multitenancy('rescuing_vif_port_id')

    def test_remove_vifs_from_node(self):
        db_utils.create_test_port(
            node_id=self.node.id, address='aa:bb:cc:dd:ee:ff',
            internal_info={driver_common.TENANT_VIF_KEY: 'test-vif-A'})
        db_utils.create_test_portgroup(
            node_id=self.node.id, address='dd:ee:ff:aa:bb:cc',
            internal_info={driver_common.TENANT_VIF_KEY: 'test-vif-B'})
        with task_manager.acquire(self.context, self.node.uuid) as task:
            network.remove_vifs_from_node(task)
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = network.get_node_vif_ids(task)
        self.assertEqual({}, result['ports'])
        self.assertEqual({}, result['portgroups'])


class TestRemoveVifsTestCase(db_base.DbTestCase):

    def setUp(self):
        super(TestRemoveVifsTestCase, self).setUp()
        self.node = object_utils.create_test_node(
            self.context,
            network_interface='flat',
            provision_state=states.DELETING)

    @mock.patch.object(neutron_common, 'unbind_neutron_port', autospec=True)
    def test_remove_vifs_from_node_failure(self, mock_unbind):
        db_utils.create_test_port(
            node_id=self.node.id, address='aa:bb:cc:dd:ee:ff',
            internal_info={driver_common.TENANT_VIF_KEY: 'test-vif-A'})
        db_utils.create_test_portgroup(
            node_id=self.node.id, address='dd:ee:ff:aa:bb:cc',
            internal_info={driver_common.TENANT_VIF_KEY: 'test-vif-B'})
        mock_unbind.side_effect = [exception.NetworkError, None]
        with task_manager.acquire(self.context, self.node.uuid) as task:
            network.remove_vifs_from_node(task)
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = network.get_node_vif_ids(task)
        self.assertEqual({}, result['ports'])
        self.assertEqual({}, result['portgroups'])
        self.assertEqual(2, mock_unbind.call_count)


class GetPortgroupByIdTestCase(db_base.DbTestCase):
    def test_portgroup_by_id(self):
        node = object_utils.create_test_node(self.context)
        portgroup = object_utils.create_test_portgroup(self.context,
                                                       node_id=node.id)
        object_utils.create_test_portgroup(self.context,
                                           node_id=node.id,
                                           uuid=uuidutils.generate_uuid(),
                                           address='00:11:22:33:44:55',
                                           name='pg2')
        with task_manager.acquire(self.context, node.uuid) as task:
            res = network.get_portgroup_by_id(task, portgroup.id)
        self.assertEqual(portgroup.id, res.id)

    def test_portgroup_by_id_no_such_portgroup(self):
        node = object_utils.create_test_node(self.context)
        object_utils.create_test_portgroup(self.context, node_id=node.id)
        with task_manager.acquire(self.context, node.uuid) as task:
            portgroup_id = 'invalid-portgroup-id'
            res = network.get_portgroup_by_id(task, portgroup_id)
        self.assertIsNone(res)


class GetPortsByPortgroupIdTestCase(db_base.DbTestCase):

    def test_ports_by_portgroup_id(self):
        node = object_utils.create_test_node(self.context)
        portgroup = object_utils.create_test_portgroup(self.context,
                                                       node_id=node.id)
        port = object_utils.create_test_port(self.context, node_id=node.id,
                                             portgroup_id=portgroup.id)
        object_utils.create_test_port(self.context, node_id=node.id,
                                      uuid=uuidutils.generate_uuid(),
                                      address='00:11:22:33:44:55')
        with task_manager.acquire(self.context, node.uuid) as task:
            res = network.get_ports_by_portgroup_id(task, portgroup.id)
        self.assertEqual([port.id], [p.id for p in res])

    def test_ports_by_portgroup_id_empty(self):
        node = object_utils.create_test_node(self.context)
        portgroup = object_utils.create_test_portgroup(self.context,
                                                       node_id=node.id)
        with task_manager.acquire(self.context, node.uuid) as task:
            res = network.get_ports_by_portgroup_id(task, portgroup.id)
        self.assertEqual([], res)


class GetPhysnetsForNodeTestCase(db_base.DbTestCase):

    def test_get_physnets_for_node_no_ports(self):
        node = object_utils.create_test_node(self.context)
        with task_manager.acquire(self.context, node.uuid) as task:
            res = network.get_physnets_for_node(task)
        self.assertEqual(set(), res)

    def test_get_physnets_for_node_excludes_None(self):
        node = object_utils.create_test_node(self.context)
        object_utils.create_test_port(self.context, node_id=node.id)
        with task_manager.acquire(self.context, node.uuid) as task:
            res = network.get_physnets_for_node(task)
        self.assertEqual(set(), res)

    def test_get_physnets_for_node_multiple_ports(self):
        node = object_utils.create_test_node(self.context)
        object_utils.create_test_port(self.context, node_id=node.id,
                                      physical_network='physnet1')
        object_utils.create_test_port(self.context, node_id=node.id,
                                      uuid=uuidutils.generate_uuid(),
                                      address='00:11:22:33:44:55',
                                      physical_network='physnet2')
        with task_manager.acquire(self.context, node.uuid) as task:
            res = network.get_physnets_for_node(task)
        self.assertEqual({'physnet1', 'physnet2'}, res)


class GetPhysnetsByPortgroupID(db_base.DbTestCase):

    def setUp(self):
        super(GetPhysnetsByPortgroupID, self).setUp()
        self.node = object_utils.create_test_node(self.context)
        self.portgroup = object_utils.create_test_portgroup(
            self.context, node_id=self.node.id)

    def _test(self, expected_result, exclude_port=None):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = network.get_physnets_by_portgroup_id(task,
                                                          self.portgroup.id,
                                                          exclude_port)
        self.assertEqual(expected_result, result)

    def test_empty(self):
        self._test(set())

    def test_one_port(self):
        object_utils.create_test_port(self.context, node_id=self.node.id,
                                      portgroup_id=self.portgroup.id,
                                      physical_network='physnet1')
        self._test({'physnet1'})

    def test_two_ports(self):
        object_utils.create_test_port(self.context, node_id=self.node.id,
                                      portgroup_id=self.portgroup.id,
                                      physical_network='physnet1')
        object_utils.create_test_port(self.context, node_id=self.node.id,
                                      uuid=uuidutils.generate_uuid(),
                                      address='00:11:22:33:44:55',
                                      portgroup_id=self.portgroup.id,
                                      physical_network='physnet1')
        self._test({'physnet1'})

    def test_exclude_port(self):
        object_utils.create_test_port(self.context, node_id=self.node.id,
                                      portgroup_id=self.portgroup.id,
                                      physical_network='physnet1')
        port2 = object_utils.create_test_port(self.context,
                                              node_id=self.node.id,
                                              uuid=uuidutils.generate_uuid(),
                                              address='00:11:22:33:44:55',
                                              portgroup_id=self.portgroup.id,
                                              physical_network='physnet2')
        self._test({'physnet1'}, port2)

    def test_exclude_port_no_id(self):
        # During port creation there may be no 'id' field.
        object_utils.create_test_port(self.context, node_id=self.node.id,
                                      portgroup_id=self.portgroup.id,
                                      physical_network='physnet1')
        port2 = object_utils.get_test_port(self.context,
                                           node_id=self.node.id,
                                           uuid=uuidutils.generate_uuid(),
                                           address='00:11:22:33:44:55',
                                           portgroup_id=self.portgroup.id,
                                           physical_network='physnet2')
        self._test({'physnet1'}, port2)

    def test_two_ports_inconsistent(self):
        object_utils.create_test_port(self.context, node_id=self.node.id,
                                      portgroup_id=self.portgroup.id,
                                      physical_network='physnet1')
        object_utils.create_test_port(self.context, node_id=self.node.id,
                                      uuid=uuidutils.generate_uuid(),
                                      address='00:11:22:33:44:55',
                                      portgroup_id=self.portgroup.id,
                                      physical_network='physnet2')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(exception.PortgroupPhysnetInconsistent,
                              network.get_physnets_by_portgroup_id,
                              task, self.portgroup.id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import socket
from unittest import mock

from oslo_config import cfg
import zeroconf

from ironic.common import exception
from ironic.common import mdns
from ironic.tests.base import TestCase

CONF = cfg.CONF


@mock.patch.object(zeroconf, 'Zeroconf', autospec=True)
class RegisterServiceTestCase(TestCase):

    def test_ok(self, mock_zc):
        zc = mdns.Zeroconf()
        zc.register_service('baremetal', 'https://127.0.0.1/baremetal')
        mock_zc.assert_called_once_with(
            interfaces=zeroconf.InterfaceChoice.All,
            ip_version=zeroconf.IPVersion.All)
        mock_zc.return_value.register_service.assert_called_once_with(mock.ANY)
        info = mock_zc.return_value.register_service.call_args[0][0]
        self.assertEqual('_openstack._tcp.local.', info.type)
        self.assertEqual('baremetal._openstack._tcp.local.', info.name)
        self.assertEqual('127.0.0.1', socket.inet_ntoa(info.addresses[0]))
        self.assertEqual({b'path': b'/baremetal'}, info.properties)

    def test_with_params(self, mock_zc):
        CONF.set_override('params', {'answer': 'none', 'foo': 'bar'},
                          group='mdns')
        zc = mdns.Zeroconf()
        zc.register_service('baremetal', 'https://127.0.0.1/baremetal',
                            params={'answer': b'42'})
        mock_zc.return_value.register_service.assert_called_once_with(mock.ANY)
        info = mock_zc.return_value.register_service.call_args[0][0]
        self.assertEqual('_openstack._tcp.local.', info.type)
        self.assertEqual('baremetal._openstack._tcp.local.', info.name)
        self.assertEqual('127.0.0.1', socket.inet_ntoa(info.addresses[0]))
        self.assertEqual({b'path': b'/baremetal',
                          b'answer': b'42',
                          b'foo': b'bar'},
                         info.properties)

    @mock.patch.object(mdns.time, 'sleep', autospec=True)
    def test_with_race(self, mock_sleep, mock_zc):
        mock_zc.return_value.register_service.side_effect = [
            zeroconf.NonUniqueNameException,
            zeroconf.NonUniqueNameException,
            zeroconf.NonUniqueNameException,
            None
        ]
        zc = mdns.Zeroconf()
        zc.register_service('baremetal', 'https://127.0.0.1/baremetal')
        mock_zc.return_value.register_service.assert_called_with(mock.ANY)
        self.assertEqual(4, mock_zc.return_value.register_service.call_count)
        mock_sleep.assert_has_calls([mock.call(i) for i in (0.1, 0.2, 0.4)])

    def test_with_interfaces(self, mock_zc):
        CONF.set_override('interfaces', ['10.0.0.1', '192.168.1.1'],
                          group='mdns')
        zc = mdns.Zeroconf()
        zc.register_service('baremetal', 'https://127.0.0.1/baremetal')
        mock_zc.assert_called_once_with(interfaces=['10.0.0.1', '192.168.1.1'],
                                        ip_version=None)
        mock_zc.return_value.register_service.assert_called_once_with(mock.ANY)
        info = mock_zc.return_value.register_service.call_args[0][0]
        self.assertEqual('_openstack._tcp.local.', info.type)
        self.assertEqual('baremetal._openstack._tcp.local.', info.name)
        self.assertEqual('127.0.0.1', socket.inet_ntoa(info.addresses[0]))
        self.assertEqual({b'path': b'/baremetal'}, info.properties)

    @mock.patch.object(mdns.time, 'sleep', autospec=True)
    def test_failure(self, mock_sleep, mock_zc):
        mock_zc.return_value.register_service.side_effect = (
            zeroconf.NonUniqueNameException
        )
        zc = mdns.Zeroconf()
        self.assertRaises(exception.ServiceRegistrationFailure,
                          zc.register_service,
                          'baremetal', 'https://127.0.0.1/baremetal')
        mock_zc.return_value.register_service.assert_called_with(mock.ANY)
        self.assertEqual(CONF.mdns.registration_attempts,
                         mock_zc.return_value.register_service.call_count)
        self.assertEqual(CONF.mdns.registration_attempts - 1,
                         mock_sleep.call_count)


class ParseEndpointTestCase(TestCase):

    def test_simple(self):
        endpoint = mdns._parse_endpoint('http://127.0.0.1')
        self.assertEqual(1, len(endpoint.addresses))
        self.assertEqual('127.0.0.1', socket.inet_ntoa(endpoint.addresses[0]))
        self.assertEqual(80, endpoint.port)
        self.assertEqual({}, endpoint.params)
        self.assertIsNone(endpoint.hostname)

    def test_simple_https(self):
        endpoint = mdns._parse_endpoint('https://127.0.0.1')
        self.assertEqual(1, len(endpoint.addresses))
        self.assertEqual('127.0.0.1', socket.inet_ntoa(endpoint.addresses[0]))
        self.assertEqual(443, endpoint.port)
        self.assertEqual({}, endpoint.params)
        self.assertIsNone(endpoint.hostname)

    def test_with_path_and_port(self):
        endpoint = mdns._parse_endpoint('http://127.0.0.1:8080/bm')
        self.assertEqual(1, len(endpoint.addresses))
        self.assertEqual('127.0.0.1', socket.inet_ntoa(endpoint.addresses[0]))
        self.assertEqual(8080, endpoint.port)
        self.assertEqual({'path': '/bm', 'protocol': 'http'}, endpoint.params)
        self.assertIsNone(endpoint.hostname)

    @mock.patch.object(socket, 'getaddrinfo', autospec=True)
    def test_resolve(self, mock_resolve):
        mock_resolve.return_value = [
            (socket.AF_INET, None, None, None, ('1.2.3.4',)),
            (socket.AF_INET6, None, None, None, ('::2', 'scope')),
        ]
        endpoint = mdns._parse_endpoint('http://example.com')
        self.assertEqual(2, len(endpoint.addresses))
        self.assertEqual('1.2.3.4', socket.inet_ntoa(endpoint.addresses[0]))
        self.assertEqual('::2', socket.inet_ntop(socket.AF_INET6,
                                                 endpoint.addresses[1]))
        self.assertEqual(80, endpoint.port)
        self.assertEqual({}, endpoint.params)
        self.assertEqual('example.com.', endpoint.hostname)
        mock_resolve.assert_called_once_with('example.com', 80, mock.ANY,
                                             socket.IPPROTO_TCP)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging
import yaml

from ironic.drivers.modules.inspector.hooks import base

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class AcceleratorsHook(base.InspectionHook):
    """Hook to set the node's accelerators property."""

    def __init__(self):
        super(AcceleratorsHook, self).__init__()
        self._known_devices = {}
        with open(CONF.inspector.known_accelerators) as f:
            self._known_devices = yaml.safe_load(f)
        self._validate_known_devices()

    def _validate_known_devices(self):
        # Do a simple check against the data source
        if (not self._known_devices
                or 'pci_devices' not in self._known_devices):
            raise RuntimeError('Could not find pci_devices in the '
                               'configuration data.')
        if not isinstance(self._known_devices['pci_devices'], list):
            raise RuntimeError('pci_devices in the configuration file should '
                               'contain a list of devices.')
        for device in self._known_devices['pci_devices']:
            if not device.get('vendor_id') or not device.get('device_id'):
                raise RuntimeError('One of the PCI devices in the '
                                   'configuration file is missing vendor_id '
                                   'or device_id.')

    def _find_accelerator(self, vendor_id, device_id):
        for dev in self._known_devices['pci_devices']:
            if (dev['vendor_id'] == vendor_id
                    and dev['device_id'] == device_id):
                return dev

    def __call__(self, task, inventory, plugin_data):
        pci_devices = plugin_data.get('pci_devices', [])

        if not pci_devices:
            LOG.warning('Unable to process accelerator devices because no PCI '
                        'device information was received from the ramdisk for '
                        'node %s.', task.node.uuid)
            return

        accelerators = []
        for pci_dev in pci_devices:
            known_device = self._find_accelerator(pci_dev['vendor_id'],
                                                  pci_dev['product_id'])
            if known_device:
                accelerator = {k: known_device[k] for k in known_device.keys()}
                accelerator.update(pci_address=pci_dev['bus'])
                accelerators.append(accelerator)

        if accelerators:
            LOG.info('Found the following accelerator devices for node %s: %s',
                     task.node.uuid, accelerators)
            task.node.set_property('accelerators', accelerators)
            task.node.save()
        else:
            LOG.info('No known accelerator devices found for node %s',
                     task.node.uuid)



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import functools
import os
import re
import sys
import time
from urllib import parse as urlparse

from keystoneauth1 import exceptions as ks_exception
import openstack
from openstack.connection import exceptions as openstack_exc
from oslo_log import log
from oslo_utils import uuidutils
import tenacity

from ironic.common import exception
from ironic.common.glance_service import service_utils
from ironic.common.i18n import _
from ironic.common import keystone
from ironic.common import swift
from ironic.common import utils
from ironic.conf import CONF

TempUrlCacheElement = collections.namedtuple('TempUrlCacheElement',
                                             ['url', 'url_expires_at'])


LOG = log.getLogger(__name__)
_GLANCE_SESSION = None


def _translate_image_exception(image_id, exc_value):
    if isinstance(exc_value, (openstack_exc.ForbiddenException)):
        return exception.ImageNotAuthorized(image_id=image_id)
    if isinstance(exc_value, openstack_exc.NotFoundException):
        return exception.ImageNotFound(image_id=image_id)
    if isinstance(exc_value, openstack_exc.BadRequestException):
        return exception.Invalid(exc_value)
    return exc_value


def check_image_service(func):
    """Creates a glance client if doesn't exists and calls the function."""
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        """Wrapper around methods calls.

        :param image_href: href that describes the location of an image
        """

        if self.client:
            return func(self, *args, **kwargs)

        global _GLANCE_SESSION
        if not _GLANCE_SESSION:
            _GLANCE_SESSION = keystone.get_session('glance')

        service_auth = keystone.get_auth('glance')

        self.endpoint = keystone.get_endpoint('glance',
                                              session=_GLANCE_SESSION,
                                              auth=service_auth)

        user_auth = None
        # NOTE(pas-ha) our ContextHook removes context.auth_token in noauth
        # case, so when ironic is in noauth but glance is not, we will not
        # enter the next if-block and use auth from [glance] config section
        if self.context.auth_token:
            user_auth = keystone.get_service_auth(self.context, self.endpoint,
                                                  service_auth)
        sess = keystone.get_session('glance',
                                    auth=user_auth or service_auth)
        conn = openstack.connection.Connection(
            session=sess,
            image_endpoint_override=self.endpoint,
            image_api_version='2')

        self.client = conn.global_request(self.context.global_id).image

        return func(self, *args, **kwargs)

    return wrapper


class GlanceImageService(object):

    # A dictionary containing cached temp URLs in namedtuples
    # in format:
    # {
    #     <image_id> : (
    #          url=<temp_url>,
    #          url_expires_at=<expiration_time>
    #     )
    # }
    _cache = {}

    def __init__(self, client=None, context=None):
        self.client = client
        self.context = context
        self.endpoint = None

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(
            exception.GlanceConnectionFailed),
        stop=utils.stop_after_retries('num_retries', group='glance'),
        wait=tenacity.wait_fixed(1),
        reraise=True
    )
    def call(self, method, *args, **kwargs):
        """Call a glance client method.

        If we get a connection error,
        retry the request according to CONF.num_retries.

        :param method: The method requested to be called.
        :param args: A list of positional arguments for the method called
        :param kwargs: A dict of keyword arguments for the method called

        :raises: GlanceConnectionFailed
        """
        try:
            return getattr(self.client, method)(*args, **kwargs)
        except openstack_exc.SDKException:
            exc_type, exc_value, exc_trace = sys.exc_info()
            new_exc = _translate_image_exception(
                args[0], exc_value)
            if isinstance(new_exc, exception.IronicException):
                # exception has been translated to a new one, raise it
                raise type(new_exc)(new_exc).with_traceback(exc_trace)
        except ks_exception.ClientException as e:
            error_msg = ("Error contacting glance endpoint "
                         "%(endpoint)s for '%(method)s'")
            LOG.exception(error_msg, {'endpoint': self.endpoint,
                                      'method': method})
            raise exception.GlanceConnectionFailed(
                endpoint=self.endpoint, reason=e)

    @check_image_service
    def show(self, image_href):
        """Returns a dict with image data for the given opaque image id.

        :param image_href: The opaque image identifier.
        :returns: A dict containing image metadata.

        :raises: ImageNotFound
        :raises: ImageUnacceptable if the image status is not active
        """
        LOG.debug("Getting image metadata from glance. Image: %s",
                  image_href)
        image_id = service_utils.parse_image_id(image_href)

        image = self.call('get_image', image_id)

        if not service_utils.is_image_active(image):
            raise exception.ImageUnacceptable(
                image_id=image_id,
                reason=_("The image is required to be in an active state."))

        if not service_utils.is_image_available(self.context, image):
            raise exception.ImageNotFound(image_id=image_id)

        base_image_meta = service_utils.translate_from_glance(image)
        return base_image_meta

    @check_image_service
    def download(self, image_href, data=None):
        """Calls out to Glance for data and writes data.

        :param image_href: The opaque image identifier.
        :param data: (Optional) File object to write data to.
        """
        image_id = service_utils.parse_image_id(image_href)

        if 'file' in CONF.glance.allowed_direct_url_schemes:
            location = self._get_location(image_id)
            url = urlparse.urlparse(location)
            if url.scheme == "file":
                with open(url.path, "r") as f:
                    filesize = os.path.getsize(f.name)
                    os.sendfile(data.fileno(), f.fileno(), 0, filesize)
                return

        image_size = 0
        image_data = None
        if data:
            image_chunks = self.call('download_image', image_id, stream=True)
            for chunk in image_chunks:
                data.write(chunk)
                image_size += len(chunk)
        else:
            image_data = self.call('download_image', image_id).content
            image_size = len(image_data)
        if image_size == 0:
            raise exception.ImageDownloadFailed(
                image_href=image_href, reason=_('image contains no data.'))
        return image_data

    def _generate_temp_url(self, path, seconds, key, method, endpoint,
                           image_id):
        """Get Swift temporary URL.

        Generates (or returns the cached one if caching is enabled) a
        temporary URL that gives unauthenticated access to the Swift object.

        :param path: The full path to the Swift object. Example:
            /v1/AUTH_account/c/o.
        :param seconds: The amount of time in seconds the temporary URL will
            be valid for.
        :param key: The secret temporary URL key set on the Swift cluster.
        :param method: A HTTP method, typically either GET or PUT, to allow for
            this temporary URL.
        :param endpoint: Endpoint URL of Swift service.
        :param image_id: UUID of a Glance image.
        :returns: temporary URL
        """

        if CONF.glance.swift_temp_url_cache_enabled:
            self._remove_expired_items_from_cache()
            if image_id in self._cache:
                return self._cache[image_id].url

        swiftapi = swift.SwiftAPI()
        path = swiftapi.generate_temp_url(
            path=path, timeout=seconds, temp_url_key=key, method=method)

        temp_url = '{endpoint_url}{url_path}'.format(
            endpoint_url=endpoint, url_path=path)

        if CONF.glance.swift_temp_url_cache_enabled:
            query = urlparse.urlparse(temp_url).query
            exp_time_str = dict(urlparse.parse_qsl(query))['temp_url_expires']
            self._cache[image_id] = TempUrlCacheElement(
                url=temp_url, url_expires_at=int(exp_time_str)
            )

        return temp_url

    def swift_temp_url(self, image_info):
        """Generate a no-auth Swift temporary URL.

        This function will generate (or return the cached one if temp URL
        cache is enabled) the temporary Swift URL using the image
        id from Glance and the config options: 'swift_endpoint_url',
        'swift_api_version', 'swift_account' and 'swift_container'.
        The temporary URL will be valid for 'swift_temp_url_duration' seconds.
        This allows Ironic to download a Glance image without passing around
        an auth_token.

        :param image_info: The return from a GET request to Glance for a
            certain image_id. Should be a dictionary, with keys like 'name' and
            'checksum'. See
            https://docs.openstack.org/glance/latest/user/glanceapi.html for
            examples.
        :returns: A signed Swift URL from which an image can be downloaded,
            without authentication.

        :raises: InvalidParameterValue if Swift config options are not set
            correctly.
        :raises: MissingParameterValue if a required parameter is not set.
        :raises: ImageUnacceptable if the image info from Glance does not
            have an image ID.
        """
        self._validate_temp_url_config()

        if ('id' not in image_info or not
                uuidutils.is_uuid_like(image_info['id'])):
            raise exception.ImageUnacceptable(_(
                'The given image info does not have a valid image id: %s')
                % image_info)

        image_id = image_info['id']

        url_fragments = {
            'api_version': CONF.glance.swift_api_version,
            'container': self._get_swift_container(image_id),
            'object_id': image_id
        }

        endpoint_url = CONF.glance.swift_endpoint_url
        if not endpoint_url:
            swift_session = swift.get_swift_session()
            try:
                endpoint_url = keystone.get_endpoint('swift',
                                                     session=swift_session)
            except exception.CatalogNotFound:
                raise exception.MissingParameterValue(_(
                    'Swift temporary URLs require a Swift endpoint URL, '
                    'but it was not found in the service catalog. You must '
                    'provide "swift_endpoint_url" as a config option.'))

        swift_account_prefix = CONF.glance.swift_account_prefix
        if swift_account_prefix and not swift_account_prefix.endswith('_'):
            swift_account_prefix = '%s_' % swift_account_prefix

        # Strip /v1/AUTH_%(tenant_id)s, if present
        endpoint_url = re.sub('/v1/%s[^/]+/?$' % swift_account_prefix, '',
                              endpoint_url)

        key = CONF.glance.swift_temp_url_key
        account = CONF.glance.swift_account
        if not account:
            swift_session = swift.get_swift_session()
            auth_ref = swift_session.auth.get_auth_ref(swift_session)
            account = '%s%s' % (swift_account_prefix, auth_ref.project_id)

        if not key:
            swift_api = swift.SwiftAPI()
            key = swift_api.get_temp_url_key()

        if not key:
            raise exception.MissingParameterValue(_(
                'Swift temporary URLs require a shared secret to be '
                'created. You must provide "swift_temp_url_key" as a '
                'config option or pre-generate the key on the project '
                'used to access Swift.'))

        url_fragments['account'] = account
        template = '/{api_version}/{account}/{container}/{object_id}'

        url_path = template.format(**url_fragments)

        return self._generate_temp_url(
            path=url_path,
            seconds=CONF.glance.swift_temp_url_duration,
            key=key,
            method='GET',
            endpoint=endpoint_url,
            image_id=image_id
        )

    def _validate_temp_url_config(self):
        """Validate the required settings for a temporary URL."""
        if (CONF.glance.swift_temp_url_duration
                < CONF.glance.swift_temp_url_expected_download_start_delay):
            raise exception.InvalidParameterValue(_(
                '"swift_temp_url_duration" must be greater than or equal to '
                '"[glance]swift_temp_url_expected_download_start_delay" '
                'option, otherwise the Swift temporary URL may expire before '
                'the download starts.'))
        seed_num_chars = CONF.glance.swift_store_multiple_containers_seed
        if (seed_num_chars is None or seed_num_chars < 0
                or seed_num_chars > 32):
            raise exception.InvalidParameterValue(_(
                "An integer value between 0 and 32 is required for"
                " swift_store_multiple_containers_seed."))

    def _get_swift_container(self, image_id):
        """Get the Swift container the image is stored in.

        Code based on: https://opendev.org/openstack/glance_store/src/commit/
        3cd690b37dc9d935445aca0998e8aec34a3e3530/glance_store/_drivers/swift/
        store.py#L725

        Returns appropriate container name depending upon value of
        ``swift_store_multiple_containers_seed``. In single-container mode,
        which is a seed value of 0, simply returns ``swift_container``.
        In multiple-container mode, returns ``swift_container`` as the
        prefix plus a suffix determined by the multiple container seed

        examples:
            single-container mode:  'glance'
            multiple-container mode: 'glance_3a1' for image uuid 3A1xxxxxxx...

        :param image_id: UUID of image
        :returns: The name of the swift container the image is stored in
        """
        seed_num_chars = CONF.glance.swift_store_multiple_containers_seed

        if seed_num_chars > 0:
            image_id = str(image_id).lower()

            num_dashes = image_id[:seed_num_chars].count('-')
            num_chars = seed_num_chars + num_dashes
            name_suffix = image_id[:num_chars]
            new_container_name = (CONF.glance.swift_container
                                  + '_' + name_suffix)
            return new_container_name
        else:
            return CONF.glance.swift_container

    def _get_location(self, image_id):
        """Get storage URL.

        Returns the direct url representing the backend storage location,
        or None if this attribute is not shown by Glance.
        """
        image_meta = self.call('get_image', image_id)

        if not service_utils.is_image_available(self.context, image_meta):
            raise exception.ImageNotFound(image_id=image_id)

        return getattr(image_meta, 'direct_url', None)

    def _remove_expired_items_from_cache(self):
        """Remove expired items from temporary URL cache

        This function removes entries that will expire before the expected
        usage time.
        """
        max_valid_time = (
            int(time.time())
            + CONF.glance.swift_temp_url_expected_download_start_delay)
        keys_to_remove = [
            k for k, v in self._cache.items()
            if (v.url_expires_at < max_valid_time)]
        for k in keys_to_remove:
            del self._cache[k]

    # TODO(TheJulia): Here because the GlanceImageService class is not based
    # upon the base image service class.
    @property
    def is_auth_set_needed(self):
        """Property to notify the caller if it needs to set authentication."""
        return False

    @property
    def transfer_verified_checksum(self):
        """The transferred artifact checksum."""
        # FIXME(TheJulia): We should look at and see if we wire
        # this up in a future change.
        return None
